@@ -10,6 +10,7 @@ import sys
 
 from snakeoil import compatibility, klass
 from snakeoil.demandload import demandload
+from snakeoil.mappings import ImmutableDict
 
 demandload(
     'itertools',
@@ -340,17 +341,20 @@ class ArgumentParser(argparse.ArgumentParser):
         self.suppress = suppress  # TODO: deprecated, drop in 0.8.0
 
         # subparser to use if none is specified on the command line and one is required
-        self._default_subparser = None
+        self.__default_subparser = None
+        # subparsers action object from calling add_subparsers()
+        self.__subparsers = None
+        # function to execute for this parser
+        self.__main_func = None
+        # arg checking function to execute for this parser
+        self.__final_check = None
 
-        # Core parser for all parent parsers, allows for separating parsing
-        # args meant for the root command with args targeted to subcommands. This
-        # allows for things such as adding conflicting options to both the root command
-        # and subcommands without causing issues in addition to helping support default
-        # subparsers.
-        self._parent = None
-        if kwds.get('parents', False):
-            self._parent = argparse.ArgumentParser(
-                add_help=False, parents=kwds.get('parents', ()))
+        # Store parent parsers allowing for separating parsing args meant for
+        # the root command with args targeted to subcommands. This enables
+        # usage such as adding conflicting options to both the root command and
+        # subcommands without causing issues in addition to helping support
+        # default subparsers.
+        self._parents = kwds.get('parents', ())
 
         # Extract the description to use, assumes first line is the short
         # description and the remainder should be used for generated docs.
@@ -442,40 +446,34 @@ class ArgumentParser(argparse.ArgumentParser):
                         where stdout is not a tty.
                     """)
 
-    @property
+    @klass.cached_property
     def subparsers(self):
         """Return the set of registered subparsers."""
-        parsers = set()
+        parsers = {}
         if self._subparsers is not None:
             for x in self._subparsers._actions:
                 if isinstance(x, argparse._SubParsersAction):
-                    parsers.update(x._name_parser_map.keys())
-        return parsers
+                    parsers.update(x._name_parser_map)
+        return ImmutableDict(parsers)
 
     def _parse_known_args(self, arg_strings, namespace):
         """Add support for using a specified, default subparser."""
-        # we don't need to determine which subparser to use
-        done = (
-            self._parent is None or  # parser has no root command
-            not arg_strings or  # no args passed in
+        skip_subparser_fallback = (
+            self.__default_subparser is None or  # no default requested
             {'-h', '--help'}.intersection(arg_strings) or  # help requested
-            arg_strings[0] in self.subparsers or  # subparser already determined
-            getattr(namespace, 'subcommand', klass._sentinel) == klass._sentinel # no subparsers exist
+            (arg_strings and arg_strings[0] in self.subparsers)  # subparser already determined
         )
 
-        if done:
-            return super(ArgumentParser, self)._parse_known_args(arg_strings, namespace)
-
-        # parse all options the parent parsers know about
-        namespace, arg_strings = self._parent._parse_known_args(arg_strings, namespace)
-
-        # add default subparser to args if one was specified
-        if self._default_subparser:
-            if self._default_subparser not in self.subparsers:
+        if not skip_subparser_fallback:
+            if self.__default_subparser not in self.subparsers:
                 raise ValueError(
                     'unknown subparser %r (available subparsers %s)' % (
-                    self._default_subparser, ', '.join(sorted(self.subparsers))))
-            arg_strings = [self._default_subparser] + arg_strings
+                    self.__default_subparser, ', '.join(sorted(self.subparsers))))
+            # parse all options the parent parsers know about
+            for parser in self._parents:
+                namespace, arg_strings = parser._parse_known_args(arg_strings, namespace)
+            # prepend the default subcommand to the current arg list
+            arg_strings = [self.__default_subparser] + arg_strings
 
         # parse the remaining args
         return super(ArgumentParser, self)._parse_known_args(arg_strings, namespace)
@@ -486,8 +484,16 @@ class ArgumentParser(argparse.ArgumentParser):
 
         args, unknown_args = self.parse_known_args(args, namespace)
 
-        # make sure prog is updated if running a subcommand
-        self.prog = getattr(args, 'prog', self.prog)
+        # make sure the correct function and prog are set if running a subcommand
+        subcmd_parser = self.subparsers.get(getattr(args, 'subcommand', None), None)
+        if subcmd_parser is not None:
+            namespace.prog = subcmd_parser.prog
+            # override the function to be run if the subcommand sets one
+            if subcmd_parser.__main_func is not None:
+                namespace.main_func = subcmd_parser.__main_func
+            # override the arg checking function if the subcommand sets one
+            if subcmd_parser.__final_check is not None:
+                namespace.final_check = subcmd_parser.__final_check
 
         if unknown_args:
             self.error('unrecognized arguments: %s' % ' '.join(unknown_args))
@@ -537,6 +543,7 @@ class ArgumentParser(argparse.ArgumentParser):
     def bind_main_func(self, functor):
         """Decorator to set a main function for the parser."""
         self.set_defaults(main_func=functor)
+        self.__main_func = functor
         # override main prog with subcommand prog
         self.set_defaults(prog=self.prog)
         return functor
@@ -558,17 +565,26 @@ class ArgumentParser(argparse.ArgumentParser):
         return f
 
     def add_subparsers(self, default=None, **kwargs):
-        self._default_subparser = default
+        # set the default subparser to use
+        self.__default_subparser = default
+
+        # If add_subparsers() has already been called return the previous
+        # object as argparse doesn't allow multiple objects of this type.
+        if self.__subparsers is not None:
+            return self.__subparsers
+
         kwargs.setdefault('title', 'subcommands')
         kwargs.setdefault('dest', 'subcommand')
         kwargs.setdefault('prog', self.prog)
         subparsers = argparse.ArgumentParser.add_subparsers(self, **kwargs)
         subparsers.required = True
+        self.__subparsers = subparsers
         return subparsers
 
     def bind_final_check(self, functor):
         """Decorator to bind a function for argument validation."""
         self.set_defaults(final_check=functor)
+        self.__final_check = functor
         return functor
 
 
